package org.apache.bcel.generic;

import org.apache.bcel.util.ByteSequence;
import java.io.IOException;
import java.io.DataOutputStream;

public abstract class BranchInstruction extends Instruction implements InstructionTargeter
{
    @Deprecated
    protected int index;
    @Deprecated
    protected InstructionHandle target;
    @Deprecated
    protected int position;
    
    static void notifyTarget(final InstructionHandle oldIh, final InstructionHandle newIh, final InstructionTargeter t) {
        if (oldIh != null) {
            oldIh.removeTargeter(t);
        }
        if (newIh != null) {
            newIh.addTargeter(t);
        }
    }
    
    BranchInstruction() {
    }
    
    protected BranchInstruction(final short opcode, final InstructionHandle target) {
        super(opcode, (short)3);
        this.setTarget(target);
    }
    
    @Override
    public boolean containsTarget(final InstructionHandle ih) {
        return this.target == ih;
    }
    
    @Override
    void dispose() {
        this.setTarget(null);
        this.index = -1;
        this.position = -1;
    }
    
    @Override
    public void dump(final DataOutputStream out) throws IOException {
        out.writeByte(super.getOpcode());
        this.index = this.getTargetOffset();
        if (!Instruction.isValidShort(this.index)) {
            throw new ClassGenException("Branch target offset too large for short: " + this.index);
        }
        out.writeShort(this.index);
    }
    
    public final int getIndex() {
        return this.index;
    }
    
    protected int getPosition() {
        return this.position;
    }
    
    public InstructionHandle getTarget() {
        return this.target;
    }
    
    protected int getTargetOffset() {
        return this.getTargetOffset(this.target);
    }
    
    protected int getTargetOffset(final InstructionHandle target) {
        if (target == null) {
            throw new ClassGenException("Target of " + super.toString(true) + " is invalid null handle");
        }
        final int t = target.getPosition();
        if (t < 0) {
            throw new ClassGenException("Invalid branch target position offset for " + super.toString(true) + ":" + t + ":" + target);
        }
        return t - this.position;
    }
    
    @Override
    protected void initFromFile(final ByteSequence bytes, final boolean wide) throws IOException {
        super.setLength(3);
        this.index = bytes.readShort();
    }
    
    protected void setIndex(final int index) {
        this.index = index;
    }
    
    protected void setPosition(final int position) {
        this.position = position;
    }
    
    public void setTarget(final InstructionHandle target) {
        notifyTarget(this.target, target, this);
        this.target = target;
    }
    
    @Override
    public String toString(final boolean verbose) {
        final String s = super.toString(verbose);
        String t = "null";
        if (this.target != null) {
            if (verbose) {
                if (this.target.getInstruction() == this) {
                    t = "<points to itself>";
                }
                else if (this.target.getInstruction() == null) {
                    t = "<null instruction!!!?>";
                }
                else {
                    t = "" + this.target.getPosition();
                }
            }
            else {
                this.index = this.target.getPosition();
                t = "" + this.index;
            }
        }
        return s + " -> " + t;
    }
    
    protected int updatePosition(final int offset, final int maxOffset) {
        this.position += offset;
        return 0;
    }
    
    @Override
    public void updateTarget(final InstructionHandle oldIh, final InstructionHandle newIh) {
        if (this.target != oldIh) {
            throw new ClassGenException("Not targeting " + oldIh + ", but " + this.target);
        }
        this.setTarget(newIh);
    }
}
